//----------------------------------------------------------
// TransformFeedBack 2016 gyabo
//----------------------------------------------------------
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <vector>

#include "win.h"
#include "wgl.h"
#include "file.h"
#include "random.h"

#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "gdi32.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "winmm.lib")

int Width = 1280;
int Height = 720;

#define TO_STRING(x)         #x
#define DATA_NUM             1024 * 1024
#define FLIP_MAX             2
#define FEEDBACK_VARYING_MAX 3
#define VERTEX_CHUNK_NUM     4
#define TEST_DATA_NUM        (DATA_NUM * VERTEX_CHUNK_NUM)

GLuint feedback_buffers[FLIP_MAX][FEEDBACK_VARYING_MAX];

void init_buffers() {
	for(int i = 0 ; i < FLIP_MAX; i++) {
		glGenBuffers(FEEDBACK_VARYING_MAX, &feedback_buffers[i][0]);
	}
	Random random;
	for(int k = 0 ; k < FLIP_MAX; k++) {
		for(int i = 0 ; i < FEEDBACK_VARYING_MAX; i++) {
			glBindBuffer(GL_ARRAY_BUFFER, feedback_buffers[k][i]);
			glBufferData(GL_ARRAY_BUFFER, VERTEX_CHUNK_NUM * sizeof(float) * TEST_DATA_NUM, 0, GL_STREAM_DRAW);
			float *d = (float *)glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
			for(int i = 0; i < TEST_DATA_NUM; i++) {
				d[i] = random.Float() * 0.3; //init
			}
			glUnmapBuffer(GL_ARRAY_BUFFER);
			glBindBuffer(GL_ARRAY_BUFFER, 0);
		}
	}
}

void AttachShader(GLuint prog, GLenum type, const char *str, const char *name = NULL) {
	GLuint shader = glCreateShader(type);
	glShaderSource(shader, 1, &str, 0);
	glCompileShader(shader);
	GLsizei buf_size = 0;
	glGetShaderiv(shader, GL_INFO_LOG_LENGTH , &buf_size);
	if(buf_size > 1) {
		std::vector<char> vbuf;
		vbuf.resize(buf_size);
		memset(&vbuf[0], 0, vbuf.size());
		glGetShaderInfoLog(shader, vbuf.size(), 0, &vbuf[0]);
		printf("[name=%s]\n%s", name, &vbuf[0]);
	}
	glAttachShader(prog, shader);
	glDeleteShader(shader);
}

GLuint CreateProgram(
	const char *vs_str,
	const char *gs_str,
	const char *fs_str,
	const char **varyings,
	int varyings_count)
{
	GLuint ret = glCreateProgram();
	if(vs_str) AttachShader(ret, GL_VERTEX_SHADER, vs_str, "VertexShader");
	if(gs_str) AttachShader(ret, GL_GEOMETRY_SHADER, gs_str, "GeometryShader");
	if(fs_str) AttachShader(ret, GL_FRAGMENT_SHADER, fs_str, "FragmentShader");

	if(varyings_count) {
		glTransformFeedbackVaryings(ret, varyings_count, varyings, GL_SEPARATE_ATTRIBS);
		printf("DEBUG %08d, CreateProgram glTransformFeedbackVaryings Proc -> glGetError=%08X\n", __LINE__, glGetError());
	}

	GLsizei buf_size = 0;
	glLinkProgram(ret);
	glGetProgramiv(ret, GL_INFO_LOG_LENGTH, &buf_size);
	if(buf_size > 1) {
		std::vector<char> vbuf;
		vbuf.resize(4096);
		memset(&vbuf[0], 0, vbuf.size());
		glGetProgramInfoLog(ret, vbuf.size(), 0, &vbuf[0]);
		printf("[glGetProgramInfoLog]\n%s\n\n", &vbuf[0]);
	}
	printf("DEBUG %08d, glLinkProgram -> glGetError=%08X\n", __LINE__, glGetError());
	return ret;
}


GLuint ctrl_prog = 0;
GLuint render_prog = 0;
GLuint query = 0;

void DeleteProgram() {
	glDeleteProgram(ctrl_prog);
	glDeleteProgram(render_prog);
}

void ReloadProgram() {
	File tfvs("tfvs.fx", "r");
	File rendervs("rendervs.fx", "r");
	File rendergs("rendergs.fx", "r");
	File renderfs("renderfs.fx", "r");

	const char* varyings[] = {
		"nextpos",
		"nextacc",
		"nextstate"
	};

	ctrl_prog = CreateProgram(
		(const char *)tfvs.Buf(),
		0,
		0,
		varyings,
		FEEDBACK_VARYING_MAX
	);

	render_prog = CreateProgram(
		(const char *)rendervs.Buf(),
		(const char *)rendergs.Buf(),
		(const char *)renderfs.Buf(),
		0,
		0
	);
	
	printf("DEBUG ctrl_prog=%d\n", ctrl_prog);
	printf("DEBUG render_prog=%d\n", render_prog);
}


void print_test_data(int index) {
	glBindBuffer(GL_ARRAY_BUFFER, feedback_buffers[index][0]);
	float *d = (float *)glMapBuffer(GL_ARRAY_BUFFER, GL_READ_ONLY);
	for(int i = 0; i < 5; i++) {
		printf("%f\n", d[i]);
	}
	glUnmapBuffer(GL_ARRAY_BUFFER);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

int main() {
	wgl_ctx ctx;
	ctx.hWnd = win_init("TFFW", Width, Height);
	wgl_init(ctx);
	
	init_buffers();
	int index = 0;
	glGenQueries(1, &query);
	while( win_proc_msg() ) {
		if(!ctrl_prog || GetAsyncKeyState(VK_F5) & 0x8000) {
			glDeleteProgram(ctrl_prog);
			glDeleteProgram(render_prog);
			ReloadProgram();
		}
		glViewport(0, 0, Width, Height);
		glClearColor(0, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		int current_index = index % FLIP_MAX;
		int next_index = (index + 1) % FLIP_MAX;

		//Ctrl(TransformFeedBack)
		{
			glUseProgram(ctrl_prog);
			glEnable(GL_RASTERIZER_DISCARD);
			
			for(int i = 0; i < FEEDBACK_VARYING_MAX; i++) {
				glEnableVertexAttribArray(i);
				glBindBuffer(GL_ARRAY_BUFFER, feedback_buffers[current_index][i]);
				glVertexAttribPointer(i, VERTEX_CHUNK_NUM, GL_FLOAT, GL_FALSE, 0, 0);
				glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, i, feedback_buffers[next_index][i]);
				glBindBuffer(GL_ARRAY_BUFFER, 0);
			}
			
			glBeginQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN, query);
			glBeginTransformFeedback(GL_POINTS);
			glDrawArrays(GL_POINTS, 0, DATA_NUM);
			glEndTransformFeedback();
			glEndQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN);
			for(int i = 0 ; i < FEEDBACK_VARYING_MAX; i++) {
				glDisableVertexAttribArray(i);
			}
			glDisable(GL_RASTERIZER_DISCARD);
			glUseProgram(0);
			glFlush();
			
			GLuint primitives;
			glGetQueryObjectuiv(query, GL_QUERY_RESULT, &primitives);
		}

		//Render
		{
			glUseProgram(render_prog);
			for(int i = 0; i < FEEDBACK_VARYING_MAX; i++) {
				glEnableVertexAttribArray(i);
				glBindBuffer(GL_ARRAY_BUFFER, feedback_buffers[next_index][i]);
				glVertexAttribPointer(i, VERTEX_CHUNK_NUM, GL_FLOAT, GL_FALSE, 0, 0);
				glBindBuffer(GL_ARRAY_BUFFER, 0);
			}
			glDrawArrays(GL_POINTS, 0, DATA_NUM);
			for(int i = 0 ; i < FEEDBACK_VARYING_MAX; i++) {
				glDisableVertexAttribArray(i);
			}
			glUseProgram(0);
		}

		index++;

		wgl_swap_buffer(ctx);
	}
	
	
	glDeleteQueries(1, &query);
	for(int i = 0 ; i < FLIP_MAX; i++) {
		glDeleteBuffers(FEEDBACK_VARYING_MAX, &feedback_buffers[i][0]);
	}
	DeleteProgram();

	wgl_term(ctx);

	return 0;
}
	

